from collections import defaultdict
import sys
from typing import Any, Callable, Iterator, List, Optional, Set, Dict

from nbformat import NotebookNode

__all__ = ['pop_recursive', 'strip_output', 'strip_zeppelin_output', 'MetadataError']


class MetadataError(Exception):
    pass


def pop_recursive(d: dict, key: str, default: Optional[NotebookNode] = None) -> NotebookNode:
    """dict.pop(key) where `key` is a `.`-delimited list of nested keys.

    >>> d = {'a': {'b': 1, 'c': 2}}
    >>> pop_recursive(d, 'a.c')
    2
    >>> d
    {'a': {'b': 1}}
    """
    if not isinstance(d, dict):
        return default
    if key in d:
        return d.pop(key, default)
    if '.' not in key:
        return default
    key_head, key_tail = key.split('.', maxsplit=1)
    if key_head in d:
        return pop_recursive(d[key_head], key=key_tail, default=default)
    return default


def _cells(nb: NotebookNode, conditionals: Callable[[NotebookNode], bool]) -> Iterator[NotebookNode]:
    """Remove cells not satisfying any conditional in conditionals and yield all other cells."""
    if hasattr(nb, 'nbformat') and nb.nbformat < 4:
        for ws in nb.worksheets:
            for conditional in conditionals:
                ws.cells = list(filter(conditional, ws.cells))
            for cell in ws.cells:
                yield cell
    else:
        for conditional in conditionals:
            nb.cells = list(filter(conditional, nb.cells))
        for cell in nb.cells:
            yield cell


def get_size(item: Any) -> int:
    """Recursively sums length of all strings in `item`"""
    if isinstance(item, str):
        return len(item)
    elif isinstance(item, list):
        return sum(get_size(elem) for elem in item)
    elif isinstance(item, dict):
        return get_size(list(item.values()))
    else:
        return len(str(item))


def determine_keep_output(cell: NotebookNode, default: bool, strip_init_cells: bool = False):
    """Given a cell, determine whether output should be kept

    Based on whether the metadata has "init_cell": true,
    "keep_output": true, or the tags contain "keep_output" """
    if 'metadata' not in cell:
        return default
    if 'init_cell' in cell.metadata:
        return bool(cell.metadata.init_cell) and not strip_init_cells

    has_keep_output_metadata = 'keep_output' in cell.metadata
    keep_output_metadata = bool(cell.metadata.get('keep_output', False))

    has_keep_output_tag = 'keep_output' in cell.metadata.get('tags', [])

    # keep_output between metadata and tags should not contradict each other
    if has_keep_output_metadata and has_keep_output_tag and not keep_output_metadata:
        raise MetadataError('cell metadata contradicts tags: `keep_output` is false, but `keep_output` in tags')

    if has_keep_output_metadata or has_keep_output_tag:
        return keep_output_metadata or has_keep_output_tag
    return default


def _zeppelin_cells(nb: dict) -> Iterator[dict]:
    for pg in nb['paragraphs']:
        yield pg


def strip_zeppelin_output(nb: dict) -> dict:
    for cell in _zeppelin_cells(nb):
        if 'results' in cell:
            cell['results'] = {}
    return nb


def match_output_type(output: Dict, output_type: str) -> bool:
    """
    Take the `output_type` string, and return whether the output matches.

    Currently, supported formats are `output_type:name` or `output_type`.

    :param output: The output dictionary from a notebook cell.
    :param output_type: User-provided string to match against.
    """

    # Check if the ':' format is used, and if so, split into output_type and name
    name = None

    if ':' in output_type:
        output_type, name = output_type.split(':')

    return output.get('output_type') == output_type and (name is None or output.get('name') == name)


def strip_output(
    nb: NotebookNode,
    keep_output: bool,
    keep_count: bool,
    keep_id: bool,
    extra_keys: List[str] = [],
    drop_empty_cells: bool = False,
    drop_tagged_cells: List[str] = [],
    strip_init_cells: bool = False,
    drop_output_types: Set[str] = None,
    keep_output_types: Set[str] = None,
    max_size: int = 0,
) -> NotebookNode:
    """
    Strip the outputs, execution count/prompt number and miscellaneous
    metadata from a notebook object, unless specified to keep either the outputs
    or counts.

    `extra_keys` could be 'metadata.foo cell.metadata.bar metadata.baz'
    """

    # Replace mutable defaults
    drop_output_types = drop_output_types or set()
    keep_output_types = keep_output_types or set()

    if keep_output is None and 'keep_output' in nb.metadata:
        keep_output = bool(nb.metadata['keep_output'])

    keys = defaultdict(list)
    for key in extra_keys:
        if '.' not in key or key.split('.')[0] not in ['cell', 'metadata']:
            sys.stderr.write(f'Ignoring invalid extra key `{key}`\n')
        else:
            namespace, subkey = key.split('.', maxsplit=1)
            keys[namespace].append(subkey)

    for field in keys['metadata']:
        pop_recursive(nb.metadata, key=field)

    conditionals = []
    # Keep cells if they have any `source` line that contains non-whitespace
    if drop_empty_cells:
        conditionals.append(lambda c: any(line.strip() for line in c.get('source', [])))
    for tag_to_drop in drop_tagged_cells:
        conditionals.append(lambda c: tag_to_drop not in c.get('metadata', {}).get('tags', []))

    for i, cell in enumerate(_cells(nb, conditionals)):
        keep_output_this_cell = determine_keep_output(cell=cell, default=keep_output, strip_init_cells=strip_init_cells)

        # Remove the outputs, unless directed otherwise
        if 'outputs' in cell:
            # Default behavior (max_size == 0) strips all outputs.
            if not keep_output_this_cell or keep_output_types:
                cell['outputs'] = [
                    output
                    for output in cell['outputs']
                    if get_size(output) <= max_size or any(match_output_type(output, ot) for ot in keep_output_types)
                ]

            # Strip the counts from the outputs that were kept if not keep_count.
            if not keep_count:
                for output in cell['outputs']:
                    if 'execution_count' in output:
                        output['execution_count'] = None

            # Remove specific output types
            if drop_output_types:
                cell['outputs'] = [
                    output
                    for output in cell['outputs']
                    if not any(match_output_type(output, ot) for ot in drop_output_types)
                ]

            # If keep_output_this_cell and keep_count, do nothing.

        # Remove the prompt_number/execution_count, unless directed otherwise
        if 'prompt_number' in cell and not keep_count:
            cell['prompt_number'] = None
        if 'execution_count' in cell and not keep_count:
            cell['execution_count'] = None
        # Replace the cell id with an incremental value that will be consistent across runs
        if 'id' in cell and not keep_id:
            cell['id'] = str(i)
        for field in keys['cell']:
            pop_recursive(cell, key=field)
    return nb



"""
Strip output from Jupyter and IPython notebooks
===============================================

Opens a notebook, strips its output, and writes the outputless version to the
original file.

Useful mainly as a git filter or pre-commit hook for users who don't want to
track output in VCS.

This does mostly the same thing as the `Clear All Output` command in the
notebook UI.

Usage
=====

Strip output from IPython / Jupyter / Zeppelin notebook (modifies the file in-place): ::

    nbstripout <file.ipynb>
    nbstripout <file.zpln>

By default, nbstripout will only modify files ending in '.ipynb' or '.zpln', to
process other files us the '-f' flag to force the application.

    nbstripout -f <file.ipynb.bak>

For using Zeppelin mode while processing files with other extensions use:
    nbstripout -m zeppelin -f <file.ext>

Use as part of a shell pipeline: ::

    cat FILE.ipynb | nbstripout > OUT.ipynb
    cat FILE.zpln | nbstripout -m zeppelin > OUT.zpln

Set up the git filter and attributes as described in the manual installation
instructions below: ::

    nbstripout --install

Set up the git filter using ``.gitattributes`` ::

    nbstripout --install --attributes .gitattributes

Set up the git filter in your global ``~/.gitconfig`` ::

    nbstripout --install --global

Set up the git filter in your system-wide ``$(prefix)/etc/gitconfig`` (most installations will require you to ``sudo``) ::

    [sudo] nbstripout --install --system

Remove the git filter and attributes: ::

    nbstripout --uninstall

Remove the git filter from your global ``~/.gitconfig`` and attributes ::

    nbstripout --uninstall --global

Remove the git filter from your system-wide ``$(prefix)/etc/gitconfig`` and attributes ::

    nbstripout --uninstall --system

Remove the git filter and attributes from ``.gitattributes``: ::

    nbstripout --uninstall --attributes .gitattributes

Check if ``nbstripout`` is installed in the current repository
(exits with code 0 if installed, 1 otherwise): ::

    nbstripout --is-installed

Print status of ``nbstripout`` installation in the current repository and
configuration summary of filter and attributes if installed
(exits with code 0 if installed, 1 otherwise): ::

    nbstripout --status

Do a dry run and only list which files would have been stripped: ::

    nbstripout --dry-run FILE.ipynb [FILE2.ipynb ...]

Print the version: ::

    nbstripout --version

Show this help page: ::

    nbstripout --help

Manual filter installation
==========================

Set up a git filter using nbstripout as follows: ::

    git config filter.nbstripout.clean '/path/to/nbstripout'
    git config filter.nbstripout.smudge cat

Create a file ``.gitattributes`` or ``.git/info/attributes`` with: ::

    *.ipynb filter=nbstripout

Apply the filter for git diff of ``*.ipynb`` files: ::

    git config diff.ipynb.textconv '/path/to/nbstripout -t'

In file ``.gitattributes`` or ``.git/info/attributes`` add: ::

    *.ipynb diff=ipynb
"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter, Namespace
import collections
import copy
import io
import json
from os import devnull, environ, makedirs, path
from pathlib import PureWindowsPath
import re
from subprocess import call, check_call, check_output, CalledProcessError, STDOUT
from typing import List, Optional
import sys
import warnings

import nbformat

from nbstripout._utils import strip_output, strip_zeppelin_output

__all__ = ['install', 'uninstall', 'status', 'main']
__version__ = '0.9.1'


INSTALL_LOCATION_LOCAL = 'local'
INSTALL_LOCATION_GLOBAL = 'global'
INSTALL_LOCATION_SYSTEM = 'system'


def _get_system_gitconfig_folder() -> str:
    try:
        git_config_output = check_output(
            ['git', 'config', '--system', '--list', '--show-origin'], universal_newlines=True, stderr=STDOUT
        ).strip()

        # If the output is empty, it means the file exists but is empty, so we cannot get the path.
        # To still get it, we're setting a temporary config parameter.
        if git_config_output == '':
            check_call(['git', 'config', '--system', 'filter.nbstripoutput.test', 'test'])
            git_config_output = check_output(
                ['git', 'config', '--system', '--list', '--show-origin'], universal_newlines=True
            ).strip()
            check_call(['git', 'config', '--system', '--unset', 'filter.nbstripoutput.test'])

        output_lines = git_config_output.split('\n')

        system_gitconfig_file_path = re.sub(r'^file:', '', output_lines[0].split('\t')[0])
    except CalledProcessError as e:
        git_config_output = e.output

        system_gitconfig_file_path = re.match(r"fatal:.*file '([^']+)'.*", git_config_output).group(1)

    return path.abspath(path.dirname(system_gitconfig_file_path))


def _get_attrfile(
    git_config: str, install_location: str = INSTALL_LOCATION_LOCAL, attrfile: Optional[str] = None
) -> str:
    if not attrfile:
        if install_location == INSTALL_LOCATION_SYSTEM:
            try:
                attrfile = check_output(git_config + ['core.attributesFile'], universal_newlines=True).strip()
            except CalledProcessError:
                config_dir = _get_system_gitconfig_folder()
                attrfile = path.join(config_dir, 'gitattributes')
        elif install_location == INSTALL_LOCATION_GLOBAL:
            try:
                attrfile = check_output(git_config + ['core.attributesFile'], universal_newlines=True).strip()
            except CalledProcessError:
                config_dir = environ.get('XDG_CONFIG_DIR', path.expanduser('~/.config'))
                attrfile = path.join(config_dir, 'git', 'attributes')
        else:
            git_dir = check_output(['git', 'rev-parse', '--git-dir'], universal_newlines=True).strip()
            attrfile = path.join(git_dir, 'info', 'attributes')

    attrfile = path.expanduser(attrfile)
    if path.dirname(attrfile):
        makedirs(path.dirname(attrfile), exist_ok=True)

    return attrfile


def _parse_size(num_str: str) -> int:
    num_str = num_str.upper()
    if num_str[-1].isdigit():
        return int(num_str)
    elif num_str[-1] == 'K':
        return int(num_str[:-1]) * (10**3)
    elif num_str[-1] == 'M':
        return int(num_str[:-1]) * (10**6)
    elif num_str[-1] == 'G':
        return int(num_str[:-1]) * (10**9)
    raise ValueError(f'Unknown size identifier {num_str[-1]}')


def install(
    git_config: str,
    install_location: str = INSTALL_LOCATION_LOCAL,
    python: Optional[str] = None,
    attrfile: Optional[str] = None,
) -> int:
    """Install the git filter and set the git attributes."""
    try:
        filepath = f'"{PureWindowsPath(python or sys.executable).as_posix()}" -m nbstripout'
        check_call(git_config + ['filter.nbstripout.clean', filepath])
        check_call(git_config + ['filter.nbstripout.smudge', 'cat'])
        check_call(git_config + ['filter.nbstripout.required', 'true'])
        check_call(git_config + ['diff.ipynb.textconv', filepath + ' -t'])
        attrfile = _get_attrfile(git_config, install_location, attrfile)
    except FileNotFoundError:
        print('Installation failed: git is not on path!', file=sys.stderr)
        return 1
    except CalledProcessError:
        print('Installation failed: not a git repository!', file=sys.stderr)
        return 1

    # Check if there is already a filter for ipynb files
    filt_exists = False
    zeppelin_filt_exists = False
    diff_exists = False

    if path.exists(attrfile):
        with open(attrfile, 'r') as f:
            attrs = f.read()

        filt_exists = '*.ipynb filter' in attrs
        zeppelin_filt_exists = '*.zpln filter' in attrs
        diff_exists = '*.ipynb diff' in attrs

        if filt_exists and diff_exists:
            return 0

    try:
        with open(attrfile, 'a', newline='') as f:
            # If the file already exists, ensure it ends with a new line
            if f.tell():
                f.write('\n')
            if not filt_exists:
                print('*.ipynb filter=nbstripout', file=f)
            if not zeppelin_filt_exists:
                print('*.zpln filter=nbstripout', file=f)
            if not diff_exists:
                print('*.ipynb diff=ipynb', file=f)
        return 0
    except PermissionError:
        print(f'Installation failed: could not write to {attrfile}', file=sys.stderr)

        if install_location == INSTALL_LOCATION_GLOBAL:
            print('Did you forget to sudo?', file=sys.stderr)

        return 1


def uninstall(git_config: str, install_location: str = INSTALL_LOCATION_LOCAL, attrfile: Optional[str] = None) -> int:
    """Uninstall the git filter and unset the git attributes."""
    try:
        call(git_config + ['--unset', 'filter.nbstripout.clean'], stdout=open(devnull, 'w'), stderr=STDOUT)
        call(git_config + ['--unset', 'filter.nbstripout.smudge'], stdout=open(devnull, 'w'), stderr=STDOUT)
        call(git_config + ['--unset', 'filter.nbstripout.required'], stdout=open(devnull, 'w'), stderr=STDOUT)
        call(git_config + ['--remove-section', 'diff.ipynb'], stdout=open(devnull, 'w'), stderr=STDOUT)
        attrfile = _get_attrfile(git_config, install_location, attrfile)
    except FileNotFoundError:
        print('Uninstall failed: git is not on path!', file=sys.stderr)
        return 1
    except CalledProcessError:
        print('Uninstall failed: not a git repository!', file=sys.stderr)
        return 1

    # Check if there is a filter for ipynb files
    if path.exists(attrfile):
        with open(attrfile, 'r+') as f:
            patterns = ('*.ipynb filter', '*.zpln filter', '*.ipynb diff')
            lines = [line for line in f if not any(line.startswith(p) for p in patterns)]
            f.seek(0)
            f.write(''.join(lines))
            f.truncate()
    return 0


def status(git_config: str, install_location: str = INSTALL_LOCATION_LOCAL, verbose: bool = False) -> int:
    """Return 0 if nbstripout is installed in the current repo, 1 otherwise"""
    try:
        if install_location == INSTALL_LOCATION_SYSTEM:
            location = 'system-wide'
        elif install_location == INSTALL_LOCATION_GLOBAL:
            location = 'globally'
        else:
            git_dir = path.dirname(
                path.abspath(check_output(['git', 'rev-parse', '--git-dir'], universal_newlines=True).strip())
            )
            location = f"in repository '{git_dir}'"

        clean = check_output(git_config + ['filter.nbstripout.clean'], universal_newlines=True).strip()
        smudge = check_output(git_config + ['filter.nbstripout.smudge'], universal_newlines=True).strip()
        diff = check_output(git_config + ['diff.ipynb.textconv'], universal_newlines=True).strip()

        if install_location in {INSTALL_LOCATION_SYSTEM, INSTALL_LOCATION_GLOBAL}:
            attrfile = _get_attrfile(git_config, install_location)
            attributes = ''
            diff_attributes = ''

            if path.exists(attrfile):
                with open(attrfile, 'r') as f:
                    attrs = f.readlines()
                attributes = ''.join(line for line in attrs if 'filter' in line).strip()
                diff_attributes = ''.join(line for line in attrs if 'diff' in line).strip()
        else:
            attributes = check_output(['git', 'check-attr', 'filter', '--', '*.ipynb'], universal_newlines=True).strip()
            diff_attributes = check_output(
                ['git', 'check-attr', 'diff', '--', '*.ipynb'], universal_newlines=True
            ).strip()

        try:
            extra_keys = check_output(git_config + ['filter.nbstripout.extrakeys'], universal_newlines=True).strip()
        except CalledProcessError:
            extra_keys = ''

        if attributes.endswith('unspecified'):
            if verbose:
                print('nbstripout is not installed', location)

            return 1

        if verbose:
            print('nbstripout is installed', location)
            print('\nFilter:')
            print('  clean =', clean)
            print('  smudge =', smudge)
            print('  diff=', diff)
            print('  extrakeys=', extra_keys)
            print('\nAttributes:\n ', attributes)
            print('\nDiff Attributes:\n ', diff_attributes)

        return 0
    except FileNotFoundError:
        print('Cannot determine status: git is not on path!', file=sys.stderr)

        return 1
    except CalledProcessError:
        if verbose and 'location' in locals():
            print('nbstripout is not installed', location)

        return 1


def process_jupyter_notebook(
    input_stream: io.IOBase,
    output_stream: io.IOBase,
    args: Namespace,
    extra_keys: List[str],
    filename: str = 'input from stdin',
) -> bool:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=UserWarning)
        nb = nbformat.read(input_stream, as_version=nbformat.NO_CONVERT)

    nb_orig = copy.deepcopy(nb)
    nb_stripped = strip_output(
        nb=nb,
        keep_output=args.keep_output,
        keep_count=args.keep_count,
        keep_id=args.keep_id,
        extra_keys=extra_keys,
        drop_empty_cells=args.drop_empty_cells,
        drop_tagged_cells=args.drop_tagged_cells.split(),
        strip_init_cells=args.strip_init_cells,
        drop_output_types=set(args.drop_output_type),
        keep_output_types=set(args.keep_output_type),
        max_size=_parse_size(args.max_size),
    )

    any_change = nb_orig != nb_stripped
    # Early exit when writing in-place and nothing changes.
    if not any_change and output_stream is input_stream:
        return any_change

    if args.dry_run:
        if any_change:
            output_stream.write(f'Dry run: would have stripped {filename}\n')
        return any_change

    if output_stream.seekable():
        output_stream.seek(0)
        output_stream.truncate()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=UserWarning)
        nbformat.write(nb_stripped, output_stream)
    try:
        output_stream.flush()
    except BrokenPipeError:
        # Receiver closed their end of the pipe after reading the data
        pass
    return any_change


def process_zeppelin_notebook(
    input_stream: io.IOBase,
    output_stream: io.IOBase,
    args: Namespace,
    extra_keys: List[str],
    filename: str = 'input from stdin',
):
    nb = json.load(input_stream, object_pairs_hook=collections.OrderedDict)
    nb_orig = copy.deepcopy(nb)
    nb_stripped = strip_zeppelin_output(nb)

    any_change = nb_orig != nb_stripped
    # Early exit when writing in-place and nothing changes.
    if not any_change and output_stream is input_stream:
        return any_change

    if args.dry_run:
        if any_change:
            output_stream.write(f'Dry run: would have stripped {filename}\n')
        return any_change

    if output_stream.seekable():
        output_stream.seek(0)
        output_stream.truncate()
    json.dump(nb_stripped, output_stream, indent=2)
    output_stream.write('\n')
    output_stream.flush()
    return any_change


def main():
    parser = ArgumentParser(epilog=__doc__, formatter_class=RawDescriptionHelpFormatter)
    task = parser.add_mutually_exclusive_group()
    task.add_argument('--dry-run', action='store_true', help='Print which notebooks would have been stripped')
    task.add_argument(
        '--install',
        action='store_true',
        help='Install nbstripout in the current repository (set up the git filter and attributes)',
    )
    task.add_argument(
        '--uninstall',
        action='store_true',
        help='Uninstall nbstripout from the current repository (remove the git filter and attributes)',
    )
    task.add_argument(
        '--is-installed', action='store_true', help='Check if nbstripout is installed in current repository'
    )
    task.add_argument(
        '--status',
        action='store_true',
        help='Print status of nbstripout installation in current repository and configuration summary if installed',
    )
    task.add_argument('--version', action='store_true', help='Print version')
    parser.add_argument(
        '--verify', action='store_true', help='Return a non-zero exit code if any files were changed, Implies --dry-run'
    )
    parser.add_argument('--keep-count', action='store_true', help='Do not strip the execution count/prompt number')
    parser.add_argument('--keep-output', action='store_true', help='Do not strip output', default=None)
    parser.add_argument(
        '--drop-output-type',
        help='Types of output cells to drop, e.g. "error" or "stream". Only has effect with --keep-output',
        nargs='+',
        default=[],
    )
    parser.add_argument(
        '--keep-output-type',
        help='Types of output cells to keep, e.g. "error" or "stream". Will take effect without --keep-output',
        nargs='+',
        default=[],
    )
    parser.add_argument(
        '--keep-id',
        action='store_true',
        help='Keep the randomly generated cell ids, which will be different after each execution.',
    )
    parser.add_argument(
        '--extra-keys',
        default='',
        help='Space separated list of extra keys to strip from metadata, e.g. metadata.foo cell.metadata.bar',
    )
    parser.add_argument(
        '--keep-metadata-keys',
        default='',
        help='Space separated list of metadata keys to keep, e.g. metadata.foo cell.metadata.bar',
    )
    parser.add_argument(
        '--drop-empty-cells',
        action='store_true',
        help='Remove cells where `source` is empty or contains only whitepace',
    )
    parser.add_argument(
        '--drop-tagged-cells', default='', help='Space separated list of cell-tags that remove an entire cell'
    )
    parser.add_argument(
        '--strip-init-cells', action='store_true', help='Remove cells with `init_cell: true` metadata (default: False)'
    )
    parser.add_argument(
        '--attributes',
        metavar='FILEPATH',
        help='Attributes file to add the filter to (in '
        'combination with --install/--uninstall), '
        'defaults to .git/info/attributes',
    )
    location = parser.add_mutually_exclusive_group()
    location.add_argument(
        '--global', dest='_global', action='store_true', help='Use global git config (default is local config)'
    )
    location.add_argument(
        '--system', dest='_system', action='store_true', help='Use system git config (default is local config)'
    )
    location.add_argument(
        '--python',
        dest='_python',
        metavar='PATH',
        help="Path to python executable to use when --install'ing (default is deduced from `sys.executable`)",
    )
    parser.add_argument(
        '--force', '-f', action='store_true', help='Strip output also from files with non ipynb extension'
    )
    parser.add_argument('--max-size', metavar='SIZE', help='Keep outputs smaller than SIZE', default='0')
    parser.add_argument(
        '--mode',
        '-m',
        default='jupyter',
        choices=['jupyter', 'zeppelin'],
        help='Specify mode between [jupyter (default) | zeppelin] (to be used in combination with -f)',
    )

    parser.add_argument('--textconv', '-t', action='store_true', help='Prints stripped files to STDOUT')

    parser.add_argument(
        '--unix-newlines',
        action='store_true',
        help='Force UNIX line endings in output (if unset, normalize to os.linesep)',
    )

    parser.add_argument('files', nargs='*', help='Files to strip output from')
    args = parser.parse_args()
    git_config = ['git', 'config']

    if args.verify and not args.dry_run:
        args.dry_run = True

    if args._system:
        git_config.append('--system')
        install_location = INSTALL_LOCATION_SYSTEM
    elif args._global:
        git_config.append('--global')
        install_location = INSTALL_LOCATION_GLOBAL
    else:
        git_config.append('--local')
        install_location = INSTALL_LOCATION_LOCAL

    if args.install:
        raise SystemExit(install(git_config, install_location, python=args._python, attrfile=args.attributes))
    if args.uninstall:
        raise SystemExit(uninstall(git_config, install_location, attrfile=args.attributes))
    if args.is_installed:
        raise SystemExit(status(git_config, install_location, verbose=False))
    if args.status:
        raise SystemExit(status(git_config, install_location, verbose=True))
    if args.version:
        print(__version__)
        raise SystemExit(0)

    extra_keys = [
        'metadata.signature',
        'metadata.widgets',
        'cell.metadata.collapsed',
        'cell.metadata.ExecuteTime',
        'cell.metadata.execution',
        'cell.metadata.heading_collapsed',
        'cell.metadata.hidden',
        'cell.metadata.scrolled',
    ]

    try:
        extra_keys.extend(
            check_output(
                (git_config if args._system or args._global else ['git', 'config']) + ['filter.nbstripout.extrakeys'],
                universal_newlines=True,
            )
            .strip()
            .split()
        )
    except (CalledProcessError, FileNotFoundError):
        pass

    extra_keys.extend(args.extra_keys.split())

    try:
        keep_metadata_keys = (
            check_output(
                (git_config if args._system or args._global else ['git', 'config'])
                + ['filter.nbstripout.keepmetadatakeys'],
                universal_newlines=True,
            )
            .strip()
            .split()
        )
    except (CalledProcessError, FileNotFoundError):
        keep_metadata_keys = []
    keep_metadata_keys.extend(args.keep_metadata_keys.split())
    extra_keys = [i for i in extra_keys if i not in keep_metadata_keys]

    # Note that we can't actually preserve newlines from the input file: nbformat implicitly converts all newlines to \n
    # and setting newline='' disables normalization of newlines on output, so the output will always use \n as newlines.
    newline = '' if args.unix_newlines else None

    # Wrap input/output stream in UTF-8 encoded text wrapper
    # https://stackoverflow.com/a/16549381
    input_stream = io.TextIOWrapper(sys.stdin.buffer, encoding='utf-8') if sys.stdin else None
    output_stream = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8', newline=newline)

    process_notebook = {'jupyter': process_jupyter_notebook, 'zeppelin': process_zeppelin_notebook}[args.mode]
    any_change = False
    for filename in args.files:
        if not (args.force or filename.endswith('.ipynb') or filename.endswith('.zpln')):
            continue

        try:
            with io.open(filename, 'r+', encoding='utf8', newline=newline) as f:
                out = output_stream if args.textconv or args.dry_run else f
                if process_notebook(
                    input_stream=f, output_stream=out, args=args, extra_keys=extra_keys, filename=filename
                ):
                    any_change = True

        except nbformat.reader.NotJSONError:
            print(f"No valid notebook detected in '{filename}'", file=sys.stderr)
            raise SystemExit(1)
        except FileNotFoundError:
            print(f"Could not strip '{filename}': file not found", file=sys.stderr)
            raise SystemExit(1)
        except Exception:
            # Ignore exceptions for non-notebook files.
            print(f"Could not strip '{filename}'", file=sys.stderr)
            raise

    if not args.files and input_stream:
        try:
            if process_notebook(input_stream, output_stream, args, extra_keys):
                any_change = True
        except nbformat.reader.NotJSONError:
            print('No valid notebook detected on stdin', file=sys.stderr)
            raise SystemExit(1)

    if args.verify and any_change:
        raise SystemExit(1)



import os
import sys
from pathlib import Path
import re
from subprocess import run, PIPE

# Note: typing.Pattern is deprecated, for removal in 3.13 in favour of re.Pattern introduced in 3.8
from typing import List, Union, Pattern

import pytest

NOTEBOOKS_FOLDER = Path('tests/e2e_notebooks')

TEST_CASES = [
    ('test_drop_empty_cells.ipynb', 'test_drop_empty_cells_dontdrop.ipynb.expected', []),
    ('test_drop_empty_cells.ipynb', 'test_drop_empty_cells.ipynb.expected', ['--drop-empty-cells']),
    ('test_drop_tagged_cells.ipynb', 'test_drop_tagged_cells_dontdrop.ipynb.expected', []),
    ('test_drop_tagged_cells.ipynb', 'test_drop_tagged_cells.ipynb.expected', ['--drop-tagged-cells=test']),
    ('test_execution_timing.ipynb', 'test_execution_timing.ipynb.expected', []),
    ('test_max_size.ipynb', 'test_max_size.ipynb.expected', ['--max-size', '50', '--keep-id']),
    ('test_max_size.ipynb', 'test_max_size.ipynb.expected_sequential_id', ['--max-size', '50']),
    ('test_empty_metadata.ipynb', 'test_empty_metadata.ipynb.expected', []),
    ('test_metadata.ipynb', 'test_metadata.ipynb.expected', []),
    (
        'test_metadata.ipynb',
        'test_metadata_extra_keys.ipynb.expected',
        ['--extra-keys', 'metadata.kernelspec metadata.language_info'],
    ),
    ('test_metadata.ipynb', 'test_metadata_keep_count.ipynb.expected', ['--keep-count']),
    ('test_metadata.ipynb', 'test_metadata_keep_output.ipynb.expected', ['--keep-output']),
    ('test_metadata.ipynb', 'test_metadata_keep_output_keep_count.ipynb.expected', ['--keep-output', '--keep-count']),
    ('test_metadata_notebook.ipynb', 'test_metadata_notebook.ipynb.expected', []),
    (
        'test_keep_metadata_keys.ipynb',
        'test_keep_metadata_keys.ipynb.expected',
        ['--keep-metadata-keys', 'cell.metadata.scrolled cell.metadata.collapsed metadata.a'],
    ),
    (
        'test_metadata_period.ipynb',
        'test_metadata_period.ipynb.expected',
        [
            '--extra-keys',
            'cell.metadata.application/vnd.databricks.v1+cell metadata.application/vnd.databricks.v1+notebook',
        ],
    ),
    ('test_strip_init_cells.ipynb', 'test_strip_init_cells.ipynb.expected', ['--strip-init-cells']),
    ('test_nbformat2.ipynb', 'test_nbformat2.ipynb.expected', []),
    ('test_nbformat45.ipynb', 'test_nbformat45.ipynb.expected', ['--keep-id']),
    ('test_nbformat45.ipynb', 'test_nbformat45.ipynb.expected_sequential_id', []),
    ('test_missing_nbformat.ipynb', 'test_missing_nbformat.ipynb.expected', []),
    ('test_unicode.ipynb', 'test_unicode.ipynb.expected', []),
    ('test_widgets.ipynb', 'test_widgets.ipynb.expected', []),
    ('test_zeppelin.zpln', 'test_zeppelin.zpln.expected', ['--mode', 'zeppelin']),
]

DRY_RUN_CASES = [
    ('test_metadata.ipynb', [], True),
    ('test_zeppelin.zpln', ['--mode', 'zeppelin'], True),
    ('test_nochange.ipynb', [], False),
]

ERR_OUTPUT_CASES = [
    (
        'test_metadata.ipynb',
        ['Ignoring invalid extra key `invalid`', 'Ignoring invalid extra key `foo.invalid`'],
        ['--extra-keys', 'invalid foo.invalid'],
    ),
    (
        'test_metadata_exception.ipynb',
        [
            re.compile(
                '.*MetadataError: cell metadata contradicts tags: `keep_output` is false, but `keep_output` in tags'
            )
        ],
        [],
    ),
    ('test_invalid_json.ipynb', ['No valid notebook detected on stdin'], []),
]


def nbstripout_exe():
    return os.environ.get('NBSTRIPOUT_EXE', 'nbstripout')


@pytest.mark.parametrize('input_file, expected_file, args', TEST_CASES)
@pytest.mark.parametrize('verify', (True, False))
def test_end_to_end_stdin(input_file: str, expected_file: str, args: List[str], verify: bool):
    with open(NOTEBOOKS_FOLDER / expected_file, mode='r') as f:
        expected = f.read()

    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        input_ = f.read()

    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        args = [nbstripout_exe()] + args
        if verify:
            args.append('--verify')
        pc = run(args, stdin=f, stdout=PIPE, universal_newlines=True)
        output = pc.stdout

    if verify:
        # When using stin, the dry run flag is disregarded.
        assert pc.returncode == (1 if input_ != expected else 0)
    else:
        assert output == expected
        assert pc.returncode == 0


@pytest.mark.parametrize('input_file, expected_file, args', TEST_CASES)
@pytest.mark.parametrize('verify', (True, False))
def test_end_to_end_file(input_file: str, expected_file: str, args: List[str], tmp_path, verify: bool):
    with open(NOTEBOOKS_FOLDER / expected_file, mode='r') as f:
        expected = f.read()

    p = tmp_path / input_file
    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        p.write_text(f.read())

    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        input_ = f.read()

    args = [nbstripout_exe(), p] + args
    if verify:
        args.append('--verify')
    pc = run(args, stdout=PIPE, universal_newlines=True)

    output = pc.stdout.strip()
    if verify:
        if expected != input_:
            assert 'Dry run: would have stripped' in output
            assert pc.returncode == 1

        # Since verify implies --dry-run, we make sure the file is not modified
        with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
            output_ = f.read()

        assert output_ == input_
    else:
        assert pc.returncode == 0
        assert not pc.stdout and p.read_text() == expected


@pytest.mark.parametrize('input_file, extra_args, any_change', DRY_RUN_CASES)
@pytest.mark.parametrize('verify', (True, False))
def test_dry_run_stdin(input_file: str, extra_args: List[str], any_change: bool, verify: bool):
    expected = 'Dry run: would have stripped input from stdin\n'

    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        args = [nbstripout_exe(), '--dry-run'] + extra_args
        if verify:
            args.append('--verify')
        pc = run(args, stdin=f, stdout=PIPE, universal_newlines=True)
        output = pc.stdout

    assert output == (expected if any_change else '')
    assert pc.returncode == (1 if verify and any_change else 0)


@pytest.mark.parametrize('input_file, extra_args, any_change', DRY_RUN_CASES)
@pytest.mark.parametrize('verify', (True, False))
def test_dry_run_args(input_file: str, extra_args: List[str], any_change: bool, verify: bool):
    expected_regex = re.compile(f'Dry run: would have stripped .*[/\\\\]{input_file}\n')
    args = [
        nbstripout_exe(),
        str(NOTEBOOKS_FOLDER / input_file),
        '--dry-run',
    ] + extra_args
    if verify:
        args.append('--verify')
    pc = run(args, stdout=PIPE, universal_newlines=True)
    output = pc.stdout

    assert expected_regex.match(output) if any_change else output == ''
    assert pc.returncode == (1 if verify and any_change else 0)


@pytest.mark.parametrize('input_file, expected_errs, extra_args', ERR_OUTPUT_CASES)
def test_make_errors(input_file: str, expected_errs: List[Union[str, Pattern]], extra_args: List[str]):
    with open(NOTEBOOKS_FOLDER / input_file, mode='r') as f:
        pc = run([nbstripout_exe(), '--dry-run'] + extra_args, stdin=f, stderr=PIPE, universal_newlines=True)
        err_output = pc.stderr

    for e in expected_errs:
        if isinstance(e, Pattern):
            assert e.search(err_output)
        else:
            assert e in err_output


def test_nochange_notebook_unchanged():
    ipynb_file = NOTEBOOKS_FOLDER / 'test_nochange.ipynb'
    ipynb_mtime_before = ipynb_file.stat().st_mtime_ns

    run([nbstripout_exe(), ipynb_file])
    ipynb_mtime_after = ipynb_file.stat().st_mtime_ns

    assert ipynb_mtime_after == ipynb_mtime_before

    zpln_file = NOTEBOOKS_FOLDER / 'test_zeppelin.zpln.expected'
    zpln_mtime_before = zpln_file.stat().st_mtime_ns

    run([nbstripout_exe(), '--force', '--mode', 'zeppelin', zpln_file])
    zpln_mtime_after = zpln_file.stat().st_mtime_ns

    assert zpln_mtime_after == zpln_mtime_before


def test_newline_behavior(tmp_path: Path):
    input_content = (NOTEBOOKS_FOLDER / 'test_drop_empty_cells.ipynb').read_bytes().replace(b'\n', b'\r\n')

    to_os_eol = tmp_path / 'should-have-os-eol.ipynb'
    to_os_eol.write_bytes(input_content)

    run([nbstripout_exe(), to_os_eol])
    if sys.platform == 'win32':
        assert b'\r\n' in to_os_eol.read_bytes()
    else:
        assert b'\r\n' not in to_os_eol.read_bytes()

    pc = run([nbstripout_exe(), '--textconv', to_os_eol], stdout=PIPE)
    if sys.platform == 'win32':
        assert b'\r\n' in pc.stdout
    else:
        assert b'\r\n' not in pc.stdout

    to_lf_eol = tmp_path / 'should-have-lf-eol.ipynb'
    to_lf_eol.write_bytes(input_content)

    run([nbstripout_exe(), '--unix-newlines', to_lf_eol])
    assert b'\r\n' not in to_lf_eol.read_bytes()

    pc = run([nbstripout_exe(), '--unix-newlines', '--textconv', to_lf_eol], stdout=PIPE)
    assert b'\r\n' not in pc.stdout



from configparser import ConfigParser
from pathlib import Path
import re
import sys

import pytest

# fix this before pytester.chdir() happens
NOTEBOOKS_FOLDER = Path('tests').absolute()


def test_install(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    assert pytester.run('nbstripout', '--is-installed').ret == 1
    pytester.run('nbstripout', '--install')
    assert pytester.run('nbstripout', '--is-installed').ret == 0

    with open('.git/info/attributes', 'r') as f:
        attr_lines = f.readlines()
    assert '*.ipynb filter=nbstripout\n' in attr_lines
    assert '*.zpln filter=nbstripout\n' in attr_lines
    assert '*.ipynb diff=ipynb\n' in attr_lines

    config = ConfigParser()
    config.read('.git/config')
    assert re.match(r'.*python.* -m nbstripout', config['filter "nbstripout"']['clean'])
    assert config['filter "nbstripout"']['required'] == 'true'
    assert config['filter "nbstripout"']['smudge'] == 'cat'
    assert re.match(r'.*python.* -m nbstripout -t', config['diff "ipynb"']['textconv'])


def test_install_different_python(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    assert pytester.run('nbstripout', '--is-installed').ret == 1
    pytester.run('nbstripout', '--install', '--python', 'DIFFERENTPYTHON')
    assert pytester.run('nbstripout', '--is-installed').ret == 0

    config = ConfigParser()
    config.read('.git/config')
    assert re.match(r'.*DIFFERENTPYTHON.* -m nbstripout', config['filter "nbstripout"']['clean'])
    assert sys.executable not in config['filter "nbstripout"']['clean']
    assert config['filter "nbstripout"']['required'] == 'true'
    assert config['filter "nbstripout"']['smudge'] == 'cat'
    assert re.match(r'.*DIFFERENTPYTHON.* -m nbstripout -t', config['diff "ipynb"']['textconv'])
    assert sys.executable not in config['diff "ipynb"']['textconv']


def test_uninstall(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    # add extra filter at the start, so we can check we don't remove it
    pytester.path.joinpath('.git/info/attributes').write_text('*.txt text')

    # do the install and verify
    pytester.run('nbstripout', '--install')
    assert pytester.run('nbstripout', '--is-installed').ret == 0

    # uninstall and verify (the actual test)
    pytester.run('nbstripout', '--uninstall')
    assert pytester.run('nbstripout', '--is-installed').ret == 1

    with open('.git/info/attributes', 'r') as f:
        attr_lines = f.readlines()
    assert '*.txt text\n' in attr_lines  # still there and not removed
    assert len(attr_lines) == 1

    config = ConfigParser()
    config.read('.git/config')
    assert 'filter "nbstripout"' not in config
    assert 'diff "ipynb"' not in config


def test_uninstall_leave_extrakeys(pytester: pytest.Pytester):
    pytester.run('git', 'init')

    # add extrakeys so we can check that we don't remove them
    pytester.run('git', 'config', 'filter.nbstripout.extrakeys', 'spam eggs')

    # check not installed
    assert pytester.run('nbstripout', '--is-installed').ret == 1

    # do the install and verify
    pytester.run('nbstripout', '--install')
    assert pytester.run('nbstripout', '--is-installed').ret == 0

    # uninstall and verify
    pytester.run('nbstripout', '--uninstall')
    assert pytester.run('nbstripout', '--is-installed').ret == 1

    # check extrakeys still exist
    r = pytester.run('git', 'config', 'filter.nbstripout.extrakeys')
    assert r.stdout.str() == 'spam eggs'


def test_status(pytester: pytest.Pytester):
    pytester.run('git', 'init')

    # status when not installed
    r = pytester.run('nbstripout', '--status')
    r.stdout.fnmatch_lines(['nbstripout is not installed in repository *'])
    assert r.ret == 1

    # do the install and verify
    pytester.run('nbstripout', '--install')
    r = pytester.run('nbstripout', '--status')
    assert r.ret == 0
    r.stdout.re_match_lines(
        r"""nbstripout is installed in repository .*
\s*
Filter:
  clean = .* -m nbstripout
  smudge = cat
  diff= .* -m nbstripout -t
  extrakeys=\s*
\s*
Attributes:
  \*.ipynb: filter: nbstripout
\s*
Diff Attributes:
  \*.ipynb: diff: ipynb
""".splitlines()
    )

    # uninstall and verify
    pytester.run('nbstripout', '--uninstall')
    r = pytester.run('nbstripout', '--status')
    r.stdout.fnmatch_lines(['nbstripout is not installed in repository *'])
    assert r.ret == 1


def test_git_diff_nodiff(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    pytester.run('git', 'config', '--local', 'filter.nbstripout.extrakeys', ' ')
    pytester.run('nbstripout', '--install')

    r = pytester.run(
        'git',
        'diff',
        '--no-index',
        '--no-ext-diff',
        '--unified=0',
        '--exit-code',
        '-a',
        '--no-prefix',
        NOTEBOOKS_FOLDER / 'test_diff.ipynb',
        NOTEBOOKS_FOLDER / 'test_diff_output.ipynb',
    )
    assert r.ret == 0
    assert not r.outlines


def test_git_diff_diff(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    pytester.run('git', 'config', '--local', 'filter.nbstripout.extrakeys', ' ')
    pytester.run('nbstripout', '--install')

    r = pytester.run(
        'git',
        'diff',
        '--no-index',
        NOTEBOOKS_FOLDER / 'test_diff.ipynb',
        NOTEBOOKS_FOLDER / 'test_diff_different_extrakeys.ipynb',
    )
    assert r.ret == 1
    r.stdout.fnmatch_lines(
        r"""index*
--- *test_diff.ipynb*
+++ *test_diff_different_extrakeys.ipynb*
@@ -6,15 +6,14 @@
    "metadata": {},
    "outputs": [],
    "source": [
-    "print(\"aou\")"
+    "print(\"aou now it is different\")"
    ]
   }
  ],
  "metadata": {
   "kernelspec": {
    "display_name": "Python 2",
-   "language": "python",
-   "name": "python2"
+   "language": "python"
   },
   "language_info": {
    "codemirror_mode": {
""".splitlines()
    )
    assert len(r.outlines) == 22  # 21 lines + new line at end


def test_git_diff_extrakeys(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    pytester.run(
        'git', 'config', '--local', 'filter.nbstripout.extrakeys', 'cell.metadata.collapsed metadata.kernelspec.name'
    )
    pytester.run('nbstripout', '--install')

    r = pytester.run(
        'git',
        'diff',
        '--no-index',
        NOTEBOOKS_FOLDER / 'test_diff.ipynb',
        NOTEBOOKS_FOLDER / 'test_diff_different_extrakeys.ipynb',
    )
    assert r.ret == 1
    r.stdout.fnmatch_lines(
        r"""index*
--- *test_diff.ipynb*
+++ *test_diff_different_extrakeys.ipynb*
@@ -6,7 +6,7 @@
    "metadata": {},
    "outputs": [],
    "source": [
-    "print(\"aou\")"
+    "print(\"aou now it is different\")"
    ]
   }
  ],
""".splitlines()
    )
    assert len(r.outlines) == 13  # 12 lines + new line at end


def test_git_diff_keepmetadatakeys(pytester: pytest.Pytester):
    pytester.run('git', 'init')
    pytester.run(
        'git', 'config', '--local', 'filter.nbstripout.keepmetadatakeys', 'cell.metadata.scrolled metadata.foo.bar'
    )
    pytester.run('nbstripout', '--install')

    r = pytester.run(
        'git',
        'diff',
        '--no-index',
        NOTEBOOKS_FOLDER / 'test_diff.ipynb',
        NOTEBOOKS_FOLDER / 'test_diff_different_extrakeys.ipynb',
    )
    assert r.ret == 1
    r.stdout.fnmatch_lines(
        r"""index*
--- *test_diff.ipynb*
+++ *test_diff_different_extrakeys.ipynb*
@@ -3,20 +3,17 @@
   {
    "cell_type": "code",
    "execution_count": null,
-   "metadata": {
-    "scrolled": true
-   },
+   "metadata": {},
    "outputs": [],
    "source": [
-    "print(\"aou\")"
+    "print(\"aou now it is different\")"
    ]
   }
  ],
""".splitlines()
    )
    assert len(r.outlines) == 28  # 12 lines + new line at end



import pytest

from nbstripout._utils import pop_recursive


def make_dict():
    return {'a': {'b': 1, 'c': 2, 'd.e': 3, 'f': {'g': 4}}}


def make_data(default=None):
    return [
        ('a.c', 2, {'a': {'b': 1, 'd.e': 3, 'f': {'g': 4}}}),
        ('a.d.e', 3, {'a': {'b': 1, 'c': 2, 'f': {'g': 4}}}),
        ('a.f', {'g': 4}, {'a': {'b': 1, 'c': 2, 'd.e': 3}}),
        ('a.f.g', 4, {'a': {'b': 1, 'c': 2, 'd.e': 3, 'f': {}}}),
        ('a', {'b': 1, 'c': 2, 'd.e': 3, 'f': {'g': 4}}, {}),
        ('notfound', default, make_dict()),
        ('a.notfound', default, make_dict()),
        ('a.b.notfound', default, make_dict()),
    ]


@pytest.fixture
def d():
    return make_dict()


@pytest.mark.parametrize(('key', 'res', 'remainder'), make_data())
def test_pop_recursive(d, key, res, remainder):
    assert pop_recursive(d, key) == res
    assert d == remainder


@pytest.mark.parametrize(('key', 'res', 'remainder'), make_data(default=0))
def test_pop_recursive_default(d, key, res, remainder):
    assert pop_recursive(d, key, default=0) == res
    assert d == remainder



from copy import deepcopy
import os
import re

import nbformat
import pytest

from nbstripout import strip_output, MetadataError

directory = os.path.dirname(__file__)


@pytest.fixture
def orig_nb():
    fname = 'test_keep_output_tags.ipynb'
    return nbformat.read(os.path.join(directory, fname), nbformat.NO_CONVERT)


@pytest.fixture
def nb_with_exception():
    fname = 'test_keep_output_tags_exception.ipynb'
    return nbformat.read(os.path.join(directory, fname), nbformat.NO_CONVERT)


def test_cells(orig_nb):
    nb_stripped = deepcopy(orig_nb)
    nb_stripped = strip_output(nb_stripped, keep_output=None, keep_count=None, keep_id=None)
    for i, cell in enumerate(nb_stripped.cells):
        if cell.cell_type == 'code' and cell.source:
            match = re.match(r'\s*#\s*(output|no_output)', cell.source)
            if match:
                # original cell should have had output.
                # If not, there's a problem with the test fixture
                assert orig_nb.cells[i].outputs

                if match.group(1) == 'output':
                    assert len(cell.outputs) > 0
                else:
                    assert len(cell.outputs) == 0


def test_exception(nb_with_exception):
    with pytest.raises(MetadataError):
        strip_output(nb_with_exception, keep_output=None, keep_count=None, keep_id=None)



from ._nbstripout import install, uninstall, status, main, __doc__ as docstring
from ._utils import pop_recursive, strip_output, MetadataError

__all__ = ['install', 'uninstall', 'status', 'main', 'pop_recursive', 'strip_output', 'MetadataError']
__doc__ = docstring



from pathlib import Path
import sys

import pytest

# fix this before pytester.chdir() happens
NOTEBOOKS_FOLDER = Path('tests').absolute()


def test_diff_with_process_substitution_nodiff(pytester: pytest.Pytester):
    if sys.platform.startswith('win'):
        pytest.skip('test requires proper bash shell')

    r = pytester.run(
        'bash',
        '-c',
        f'diff <( nbstripout -t {NOTEBOOKS_FOLDER / "test_diff.ipynb"} ) <( nbstripout -t {NOTEBOOKS_FOLDER / "test_diff_output.ipynb"} )',
    )
    assert not r.outlines
    assert r.ret == 0


def test_diff_with_process_substitution_diff(pytester: pytest.Pytester):
    if sys.platform.startswith('win'):
        pytest.skip('test requires proper bash shell')

    r = pytester.run(
        'bash',
        '-c',
        f'diff <( nbstripout -t {NOTEBOOKS_FOLDER / "test_diff.ipynb"} ) <( nbstripout -t {NOTEBOOKS_FOLDER / "test_diff_different.ipynb"} )',
    )
    r.stdout.re_match_lines(
        r"""(.*)
<     "print(\"aou\")"
---
(.*\"print\(\\\"aou now it is different\\\"\)\")
""".splitlines()
    )
    assert r.ret == 1



import os
from copy import deepcopy

import nbformat
import pytest

from nbstripout import strip_output

directory = os.path.dirname(__file__)


@pytest.fixture
def orig_nb():
    # Grab the original notebook
    fname = 'test_output_types.ipynb'
    return nbformat.read(os.path.join(directory, fname), nbformat.NO_CONVERT)


def test_drop_errors(orig_nb):
    """
    Confirm that --drop-output-types works as expected,
    when asking just to drop `error` outputs.
    """
    nb_stripped = strip_output(
        deepcopy(orig_nb), keep_output=True, keep_count=False, keep_id=False, drop_output_types={'error'}
    )

    # No outputs in the markdown
    assert not hasattr(nb_stripped.cells[0], 'outputs')

    # Original cell should have 3 outputs, with the last being error
    assert len(orig_nb.cells[1].outputs) == 3
    assert orig_nb.cells[1].outputs[2]['output_type'] == 'error'

    # Second cell should have a stdout stream, stderr stream and an error
    stripped_output_1 = nb_stripped.cells[1].outputs[0]
    stripped_output_2 = nb_stripped.cells[1].outputs[1]
    assert len(nb_stripped.cells[1].outputs) == 2
    assert stripped_output_1['output_type'] == 'stream'
    assert stripped_output_2['output_type'] == 'stream'
    assert stripped_output_1['name'] == 'stdout'
    assert stripped_output_2['name'] == 'stderr'

    # Third cell should have an execution output
    assert len(nb_stripped.cells[2].outputs) == 1
    assert nb_stripped.cells[2].outputs[0]['output_type'] == 'execute_result'


def test_keep_output(orig_nb):
    """
    Confirm that --keep-output-types works as expected,
    dropping all but `execute_result` outputs from the notebook.
    """
    nb_stripped = strip_output(
        deepcopy(orig_nb), keep_output=True, keep_count=False, keep_id=False, keep_output_types={'execute_result'}
    )

    # No outputs in the markdown
    assert not hasattr(nb_stripped.cells[0], 'outputs')

    # Original cell should have 3 outputs, with the last being error
    assert len(orig_nb.cells[1].outputs) == 3
    assert orig_nb.cells[1].outputs[2]['output_type'] == 'error'

    # All outputs should be stripped in the second cell
    assert len(nb_stripped.cells[1].outputs) == 0

    # Third cell should have an execution output
    assert len(nb_stripped.cells[2].outputs) == 1


def test_output_format_tags(orig_nb):
    """
    Confirm that both <output_type>:<name> and <output_type> formats work.
    """
    nb_stripped = strip_output(
        deepcopy(orig_nb),
        keep_output=False,
        keep_count=False,
        keep_id=False,
        keep_output_types={'stream:stdout', 'execute_result'},
    )

    # No outputs in the markdown
    assert not hasattr(nb_stripped.cells[0], 'outputs')

    # Stripping all but stdout should leave only the print statement
    assert len(orig_nb.cells[1].outputs) == 3
    assert len(nb_stripped.cells[1].outputs) == 1

    # Third cell should have only the execute_result
    assert len(nb_stripped.cells[2].outputs) == 1
